import torch
from torch import pi, sqrt
from caskade import Module, forward, Param
from caustics.light.base import Source
import numpy as np
from numpy.polynomial.legendre import leggauss
from pykeops.torch import LazyTensor
from torch.nn.functional import conv2d, avg_pool2d
from functools import lru_cache
import math


@lru_cache(maxsize=None)
def _leggauss_const(n, dtype, device):
    x_np, w_np = np.polynomial.legendre.leggauss(n)
    return (torch.as_tensor(x_np, dtype=dtype, device = device),
            torch.as_tensor(w_np, dtype=dtype, device = device))

# 2.  Pure-Torch mapping keeps autograd alive and avoids graph breaks.
def leggauss_interval(n, t_low, t_high, device=None, dtype=None):
    x0, w0 = _leggauss_const(n, dtype, device)
    half_const = torch.tensor(0.5, dtype=dtype, device=device)

    half = half_const * (t_high - t_low)
    mid  = half_const * (t_high + t_low)

    # allow t_low / t_high to be batched – add a dim for broadcasting
    x = half.unsqueeze(-1) * x0 + mid.unsqueeze(-1)
    w = half.unsqueeze(-1) * w0
    return x, w


def transform_DE(t):
    """
    Double-exponential transform:
      u = exp((π/2) * sinh(t)),
      du/dt = (π/2)*cosh(t)*u.
    """
    u = torch.exp((np.pi/2.0)*torch.sinh(t))
    du_dt = (np.pi/2.0)*torch.cosh(t)*u
    return u, du_dt
    

def interpolate_velocity(R_grid: torch.Tensor,
                         R_map : torch.Tensor,
                         v_grid: torch.Tensor) -> torch.Tensor:
    """
    1-D linear interpolation on an arbitrary monotonic grid.
    Any value outside [R_grid[0], R_grid[-1]] is clamped to the edges.
    Works on CUDA tensors, keeps gradients, avoids out-of-bounds.
    """
    # 1. Clamp the query points to the grid range
    R_clamp = R_map.clamp(min=R_grid[0], max=R_grid[-1])

    # 2. Locate the interval: first index such that R_grid[idx_hi] ≥ R_clamp
    idx_hi = torch.searchsorted(R_grid, R_clamp, right=False)

    #   For values equal to R_grid[-1] we still get idx_hi == len(R_grid)
    idx_hi = idx_hi.clamp(max=R_grid.numel() - 1)

    # 3. Lower neighbour
    idx_lo = (idx_hi - 1).clamp(min=0)

    # 4. Gather the two bracketing points
    R_lo, R_hi = R_grid[idx_lo], R_grid[idx_hi]
    v_lo, v_hi = v_grid[idx_lo], v_grid[idx_hi]

    # 5. Linear weight (when R_lo == R_hi, weight → 0)
    w = torch.where(
        R_hi == R_lo,
        torch.zeros_like(R_lo),
        (R_clamp - R_lo) / (R_hi - R_lo)
    )

    return v_lo + w * (v_hi - v_lo)


class MGEVelocityIntr(Module):
    """
    Identical setup to `ThinMGEVelocity`, but now we use the intrinsic q directly.
    """
    def __init__(self, N_components: int, device, dtype, quad_points=128, radius_res = 4096, variable_M_to_L = False, soft = 0.0, G=0.004301):
        """
        Soft: softening length in parsecs
        """
        super().__init__("MGEVelocityIntr")
        self.device = device
        self.dtype  = dtype
        
        self.N_components = N_components
        
        # Same parameter definitions
        self.surf   = Param("surf",   shape=(N_components,))
        self.sigma  = Param("sigma",  shape=(N_components,))
        self.qintr   = Param("qintr",   shape=(N_components,))
        if variable_M_to_L:
            self.M_to_L = Param("M_to_L", shape=(N_components,))
        else:
            self.M_to_L = Param("M_to_L", shape=())
        
        self.m_bh  = Param("m_bh",  shape=())
        self.quad_points = quad_points
        self.radius_res = radius_res

        self.soft = torch.tensor(soft, device=device, dtype=dtype)
        self.G = torch.tensor(G, device=device, dtype=dtype)
        self.inc = Param("inc",   shape=())

    def radial_velocity(self, R_flat,
                 surf, sigma, qintr, M_to_L,
                 inc, m_bh):
        """
        Compute the rotational velocity at radii R_flat, but use a
        double-exponential transform from [0,1] -> (0,∞).
        """
        # --- Type-Safe Constants Definition ---
        # Define EVERY float literal as a tensor to prevent silent promotion.
        _p5 = torch.tensor(0.5, device=self.device, dtype=self.dtype)
        _1 = torch.tensor(1.0, device=self.device, dtype=self.dtype)
        _2 = torch.tensor(2.0, device=self.device, dtype=self.dtype)
        _10 = torch.tensor(10.0, device=self.device, dtype=self.dtype)
        _pi = torch.tensor(np.pi, device=self.device, dtype=self.dtype)
        _1e_7 = torch.tensor(1e-7, device=self.device, dtype=self.dtype)
        _1e3 = torch.tensor(1e3, device=self.device, dtype=self.dtype)
        _neg_1p5 = torch.tensor(-1.5, device=self.device, dtype=self.dtype)
        # --- End Constants Definition ---

        sqrt_2pi = torch.sqrt(_2 * _pi)
        qobs = torch.sqrt(qintr**2 * (torch.sin(inc))**2 + (torch.cos(inc))**2)
        mass_density = surf * M_to_L * qobs / (qintr * sigma * sqrt_2pi)

        N_points = R_flat.shape[0]

        # Scale by median sigma
        scale = sigma.quantile(q=0.5)
        sigma_sc = sigma / scale
        R_sc = R_flat / scale
        soft_sc = self.soft / scale

        mds = sigma_sc.quantile(q=0.5)
        mxs = torch.max(sigma_sc)

        xlim = (torch.arcsinh(torch.log(_1e_7 * mds) * _2 / _pi),
                torch.arcsinh(torch.log(_1e3 * mxs) * _2 / _pi))

        # --- Gauss–Legendre on [0,1] ---
        lo, hi = xlim
        t_1d, w_1d = leggauss_interval(self.quad_points, lo, hi, device=self.device, dtype=self.dtype)

        # --- Double-exponential transform t->u in (0,∞) ---
        u_1d, du_1d = transform_DE(t_1d)

        R_i = R_sc.view(-1, 1, 1)                     # (N,1,1)
        u_j = u_1d.view(1, -1, 1)                    # (1,Q,1)
        w_j = w_1d.view(1, -1, 1)                    # (1,Q,1)
        du_j = du_1d.view(1, -1, 1)                    # (1,Q,1)

        sigma_mat = sigma_sc.view(1, 1, -1)         # (1,1,C)
        qintr_mat = qintr.view(1, 1, -1)           # (1,1,C)
        mass_den_mat = mass_density.view(1, 1, -1)     # (1,1,C)

        # ---- kernel -----------------------------------------------------------------
        one_plus = _1 + u_j
        exp_val = torch.exp(-_p5 * R_i.pow(_2) /
                             (sigma_mat.pow(_2) * one_plus))

        denom = one_plus.pow(_2) * torch.sqrt(qintr_mat.pow(_2) + u_j)

        term = (qintr_mat * mass_den_mat * exp_val) / denom
        weighted = term * du_j * w_j

        # ---- quadrature & component sums -------------------------------------------
        integral_val = weighted.sum(dim=1).sum(dim=1)

        # ---- finish exactly as before ----------------------------------------------
        vc2_mge_factor = _2 * _pi * self.G * (scale**_2)
        vc2_mge = vc2_mge_factor * integral_val

        vc2_bh = self.G * _10**m_bh / scale * (R_sc**_2 + soft_sc**_2).pow(_neg_1p5)

        v_rot_flat = R_sc * torch.sqrt(vc2_mge + vc2_bh)

        return v_rot_flat
        
    @forward
    def velocity(
        self,
        R_map,                           # 2-D tensor [H,W]  (pc)
        surf=None, sigma=None, qintr=None, M_to_L=None, inc = None, m_bh=None
    ):
        """
        Returns v_rot(R) for every pixel in the sky plane.
        """
        Rmin = torch.as_tensor(self.soft, dtype=self.dtype, device=self.device)
        Rmax = R_map.max()

        # 1-D lookup table (same as before)
        R_grid = torch.logspace(
            torch.log10(Rmin),
            torch.log10(Rmax),
            self.radius_res,
            device=self.device,
            dtype=self.dtype,
        )
        v_grid = self.radial_velocity(
            R_grid, surf, sigma, qintr, M_to_L, inc, m_bh
        )

        # interpolate onto the pixel-by-pixel radii
        v_abs = interpolate_velocity(R_grid, R_map, v_grid)      # (H,W)

        return v_abs

class Nuker_MGE(Module):
    def __init__(self, N_MGE_components: int, Nuker_NN, NN_dtype, distance, r_min, r_max, soft, device, dtype, quad_points=128):
        super().__init__("NukerMGE")
        self.N_components = N_MGE_components
        self.soft = soft
        self.MGE = MGEVelocityIntr(self.N_components, soft = soft, quad_points = quad_points, dtype = dtype, device = device)
        self.MGE.surf = torch.ones((self.N_components), device = device).to(dtype = dtype)
        self.MGE.sigma = torch.ones((self.N_components), device = device).to(dtype = dtype)
        self.MGE.qintr = torch.ones((self.N_components), device = device).to(dtype = dtype)
        self.MGE.M_to_L = torch.tensor([1.0], dtype = dtype, device = device)
        self.NN = Nuker_NN

        inner_slope=torch.tensor([3.0], device = device, dtype = dtype)
        outer_slope=torch.tensor([3.0], device = device, dtype = dtype)
        low_Gauss=torch.log10(r_min/torch.sqrt(inner_slope))
        high_Gauss=torch.log10(r_max/torch.sqrt(outer_slope))
        dx=(high_Gauss-low_Gauss)/self.N_components
        
        # --- SOLUTION ---
        # Ensure all scalars are tensors of the correct dtype before the calculation
        distance_t = torch.tensor(distance, device=device, dtype=dtype)
        pi_t = torch.tensor(np.pi, device=device, dtype=dtype)
        
        self.sigma = (distance_t * (pi_t / 0.648)) * 10**(low_Gauss + (0.5 + torch.arange(self.N_components, device=device, dtype=dtype)) * dx)
        
        self.inc   = Param("inc",   shape=())
        self.qintr = Param("qintr", shape=())
        self.qintr_shaper = torch.ones((self.N_components), device = device).to(dtype = dtype)
        self.m_bh  = Param("m_bh",  shape=())
        self.MGE.inc = self.inc
        self.MGE.m_bh = self.m_bh

        self.alpha = Param("alpha", shape=(1, ))
        self.gmb = Param("gamma_minus_beta", shape=(1, ))
        self.gamma = Param("gamma", shape=(1, ))
        self.r_b = Param("break_r", shape = ())
        self.I_b = Param("intensity_r_b", shape = ())
        self.dtype = dtype
        self.NN_dtype = NN_dtype
    
    def symexp(self, y, linthresh=1e-12, base=10.0):
        # --- SOLUTION ---
        # Create tensor constants that match the input tensor's properties
        linthresh_t = torch.tensor(linthresh, device=y.device, dtype=y.dtype)
        base_t = torch.tensor(base, device=y.device, dtype=y.dtype)
        one_t = torch.tensor(1.0, device=y.device, dtype=y.dtype)
    
        return torch.sign(y) * linthresh_t * (base_t**torch.abs(y) - one_t)

    @forward
    def velocity(self, R_flat,
                 inc=None, qintr=None, m_bh=None,
                 alpha = None, gmb = None, gamma = None, r_b = None, I_b = None,
                 G=0.004301):
        device = R_flat.device
        dtype  = R_flat.dtype
        beta = gamma - gmb

        NN_input = torch.cat([alpha, beta, gamma])#.to(self.NN_dtype)
        NN_output_transformed = self.NN.forward(NN_input)#.to(self.dtype)
        NN_output = self.symexp(NN_output_transformed)
        
        surf = NN_output*10**I_b
        MGE_sigma = self.sigma*r_b
        v_rot = self.MGE.velocity(R_map = R_flat, surf = surf, sigma = MGE_sigma, qintr = qintr*self.qintr_shaper)
        return v_rot


######################################################################################################
# Older MGE implementations (3D Rs, qobs, etc)
######################################################################################################

class ThinMGEVelocity(Module):
    """
    Identical parameters to `MGEVelocity`, but the `forward` method now
    takes *only* a radius map R_map instead of (x,y,z) grids.
    """
    def __init__(self, N_components: int, device, dtype, quad_points=128, radius_res = 4096, variable_M_to_L = False, soft = 0.0, G=0.004301):
        """
        Soft: softening length in parsecs
        """
        super().__init__("MGEVelocity")
        self.device = device
        self.dtype  = dtype
        
        self.N_components = N_components
        
        # Same parameter definitions
        self.surf   = Param("surf",   shape=(N_components,))
        self.sigma  = Param("sigma",  shape=(N_components,))
        self.qobs   = Param("qobs",   shape=(N_components,))
        if variable_M_to_L:
            self.M_to_L = Param("M_to_L", shape=(N_components,))
        else:
            self.M_to_L = Param("M_to_L", shape=())
        
        self.inc   = Param("inc",   shape=())
        self.m_bh  = Param("m_bh",  shape=())
        self.quad_points = quad_points
        self.radius_res = radius_res

        self.soft = soft
        self.G = G

    def radial_velocity(self, R_flat,
                 surf, sigma, qobs, M_to_L,
                 inc, m_bh):
        """
        Compute the rotational velocity at radii R_flat, but use a
        double-exponential transform from [0,1] -> (0,∞).
        """
        
        # --- geometry & mass density the same as your original ---
        cos_inc = torch.cos(inc)
        sin_inc = torch.sin(inc)
        q_arg = qobs**2 - cos_inc**2
        if torch.any(q_arg <= 0.0):
            raise ValueError("Inclination too low for deprojection")
        q_intr = torch.sqrt(q_arg) / sin_inc
        
        sqrt_2pi = np.sqrt(2.0*np.pi)
        mass_density = surf * M_to_L * qobs / (q_intr * sigma * sqrt_2pi)
        
        N_points = R_flat.shape[0]
        
        # Scale by median sigma
        scale = sigma.quantile(q = 0.5)
        sigma_sc = sigma / scale
        R_sc = R_flat / scale
        soft_sc = self.soft / scale

        mds = sigma_sc.quantile(q = 0.5)
        mxs = torch.max(sigma_sc)
        xlim = (torch.arcsinh(torch.log(1e-7 * mds)*2/np.pi),
            torch.arcsinh(torch.log(1e3  * mxs)*2/np.pi))
        
        # --- Gauss–Legendre on [0,1] ---
        lo, hi = xlim
        t_1d, w_1d = leggauss_interval(self.quad_points, lo, hi, device = self.device, dtype = self.dtype)

        
        # --- Double-exponential transform t->u in (0,∞) ---
        u_1d, du_1d = transform_DE(t_1d)
        
        R_i  = R_sc.view(-1, 1, 1)                     # (N,1,1)
        u_j  = u_1d.view(1,  -1, 1)                    # (1,Q,1)torch.linspace(-1, 1, nx, device=device, dtype=dtype) * pixelscale * (nx - 1) / 2
        w_j  = w_1d.view(1,  -1, 1)                    # (1,Q,1)
        du_j = du_1d.view(1, -1, 1)                    # (1,Q,1)
        
        sigma_mat    = sigma_sc.view(1, 1, -1)         # (1,1,C)
        q_intr_mat   = q_intr.view(1, 1, -1)           # (1,1,C)
        mass_den_mat = mass_density.view(1, 1, -1)     # (1,1,C)
        
        # ---- kernel -----------------------------------------------------------------
        one_plus = 1.0 + u_j                           # (1,Q,1)
        exp_val  = torch.exp(-0.5 * R_i.pow(2) /
                             (sigma_mat.pow(2) * one_plus))          # (N,Q,C)
        
        denom    = one_plus.pow(2) * torch.sqrt(q_intr_mat.pow(2) + u_j)
        
        term      = (q_intr_mat * mass_den_mat * exp_val) / denom    # (N,Q,C)
        weighted  = term * du_j * w_j                                # (N,Q,C)
        
        # ---- quadrature & component sums -------------------------------------------
        integral_val = weighted.sum(dim=1).sum(dim=1)   # (N,)
        
        # ---- finish exactly as before ----------------------------------------------
        vc2_mge_factor = 2.0 * np.pi * self.G * (scale**2)
        vc2_mge = vc2_mge_factor * integral_val
        
        vc2_bh = self.G * 10**m_bh / scale * (R_sc**2 + soft_sc**2).pow(-1.5)
        
        v_rot_flat = R_sc * torch.sqrt(vc2_mge + vc2_bh)   # (N,)
        
        return v_rot_flat
        
    @forward
    def velocity(
        self,
        R_map,                           # 2-D tensor [H,W]  (pc)
        surf=None, sigma=None, qobs=None, M_to_L=None,
        inc=None, m_bh=None,
    ):
        """
        Returns v_rot(R) for every pixel in the sky plane.
        """

        Rmin = torch.as_tensor(self.soft, dtype=self.dtype, device=self.device)
        Rmax = R_map.max()

        # 1-D lookup table (same as before)
        R_grid = torch.logspace(
            torch.log10(Rmin),
            torch.log10(Rmax),
            self.radius_res,
            device=self.device,
            dtype=self.dtype,
        )
        v_grid = self.radial_velocity(
            R_grid, surf, sigma, qobs, M_to_L, inc, m_bh
        )

        # interpolate onto the pixel-by-pixel radii
        v_abs = interpolate_velocity(R_grid, R_map, v_grid)      # (H,W)

        return v_abs


class MGEVelocity(Module):
    def __init__(self, N_components: int, device, dtype, quad_points=128, radius_res = 4096, variable_M_to_L = False, soft = 0.0, G=0.004301):
        """
        Soft: softening length in parsecs
        """
        super().__init__("MGEVelocity")
        self.device = device
        self.dtype  = dtype
        
        self.N_components = N_components
        
        # Same parameter definitions
        self.surf   = Param("surf",   shape=(N_components,))
        self.sigma  = Param("sigma",  shape=(N_components,))
        self.qobs   = Param("qobs",   shape=(N_components,))
        if variable_M_to_L:
            self.M_to_L = Param("M_to_L", shape=(N_components,))
        else:
            self.M_to_L = Param("M_to_L", shape=())
        
        self.inc   = Param("inc",   shape=())
        self.m_bh  = Param("m_bh",  shape=())
        self.quad_points = quad_points
        self.radius_res = radius_res

        self.soft = soft
        self.G = G

    def radial_velocity(self, R_flat,
                 surf, sigma, qobs, M_to_L,
                 inc, m_bh):
        """
        Compute the rotational velocity at radii R_flat, but use a
        double-exponential transform from [0,1] -> (0,∞).
        """
        
        # --- geometry & mass density the same as your original ---
        cos_inc = torch.cos(inc)
        sin_inc = torch.sin(inc)
        q_arg = qobs**2 - cos_inc**2
        if torch.any(q_arg <= 0.0):
            raise ValueError("Inclination too low for deprojection")
        q_intr = torch.sqrt(q_arg) / sin_inc
        
        sqrt_2pi = np.sqrt(2.0*np.pi)
        mass_density = surf * M_to_L * qobs / (q_intr * sigma * sqrt_2pi)
        
        N_points = R_flat.shape[0]
        
        # Scale by median sigma
        scale = sigma.quantile(q = 0.5)
        sigma_sc = sigma / scale
        R_sc = R_flat / scale
        soft_sc = self.soft / scale

        mds = sigma_sc.quantile(q = 0.5)
        mxs = torch.max(sigma_sc)
        xlim = torch.tensor([
            torch.arcsinh(torch.log(1e-7 * mds)*2/np.pi),
            torch.arcsinh(torch.log(1e3  * mxs)*2/np.pi)
        ])
        
        # --- Gauss–Legendre on [0,1] ---
        t_1d, w_1d = leggauss_interval(self.quad_points, xlim[0].item(), xlim[1].item(), self.device, self.dtype)
        
        # --- Double-exponential transform t->u in (0,∞) ---
        u_1d, du_1d = transform_DE(t_1d)
        
        R_i  = R_sc.view(-1, 1, 1)                     # (N,1,1)
        u_j  = u_1d.view(1,  -1, 1)                    # (1,Q,1)torch.linspace(-1, 1, nx, device=device, dtype=dtype) * pixelscale * (nx - 1) / 2
        w_j  = w_1d.view(1,  -1, 1)                    # (1,Q,1)
        du_j = du_1d.view(1, -1, 1)                    # (1,Q,1)
        
        sigma_mat    = sigma_sc.view(1, 1, -1)         # (1,1,C)
        q_intr_mat   = q_intr.view(1, 1, -1)           # (1,1,C)
        mass_den_mat = mass_density.view(1, 1, -1)     # (1,1,C)
        
        # ---- kernel -----------------------------------------------------------------
        one_plus = 1.0 + u_j                           # (1,Q,1)
        exp_val  = torch.exp(-0.5 * R_i.pow(2) /
                             (sigma_mat.pow(2) * one_plus))          # (N,Q,C)
        
        denom    = one_plus.pow(2) * torch.sqrt(q_intr_mat.pow(2) + u_j)
        
        term      = (q_intr_mat * mass_den_mat * exp_val) / denom    # (N,Q,C)
        weighted  = term * du_j * w_j                                # (N,Q,C)
        
        # ---- quadrature & component sums -------------------------------------------
        integral_val = weighted.sum(dim=1).sum(dim=1)   # (N,)
        
        # ---- finish exactly as before ----------------------------------------------
        vc2_mge_factor = 2.0 * np.pi * self.G * (scale**2)
        vc2_mge = vc2_mge_factor * integral_val
        
        vc2_bh = self.G * 10**m_bh / scale * (R_sc**2 + soft_sc**2).pow(-1.5)
        
        v_rot_flat = R_sc * torch.sqrt(vc2_mge + vc2_bh)   # (N,)
        
        return v_rot_flat

    @forward
    def velocity(self, rot_x, rot_y, rot_z,
                 surf=None, sigma=None, qobs=None, M_to_L=None,
                 inc=None, m_bh=None):
        """
        Compute the rotational velocity at points (x, y, z), but use a
        double-exponential transform from [0,1] -> (0,∞).
        """
        # cylindrical radii in the **galaxy** frame
        R_map = torch.sqrt(rot_x**2 + rot_y**2)
        Rmin = self.soft
        Rmax = R_map.max()

        R_grid = torch.logspace(np.log10(Rmin), np.log10(Rmax), self.radius_res,
                        device=self.device, dtype=self.dtype)
        
        v_grid = self.radial_velocity(R_grid, surf, sigma, qobs, M_to_L, inc, m_bh)

        v_abs = interpolate_velocity(R_grid, R_map, v_grid)

        return v_abs
        
       











import math, torch
from torch.quasirandom import SobolEngine
from torch.utils.data import DataLoader, TensorDataset
from tqdm import tqdm

# ──────────────────────────────────────────────────────────────
# 1. Sobol' sampler for uniform priors
# ──────────────────────────────────────────────────────────────
def sobol_sample(param_bounds,           # list/tuple/array of shape (D, 2)
                 n_samples, 
                 *, 
                 dtype=torch.float64,
                 device="cpu",
                 scramble=True,
                 seed=None):
    """
    Draw `n_samples` Sobol' points within hyper‑rectangular bounds.
    
    Parameters
    ----------
    param_bounds : sequence[(low, high), …] or (D, 2) tensor
        Uniform prior box for each parameter.
    n_samples    : int
        Number of Sobol' points to generate.
    dtype, device, scramble, seed : usual Torch options.
    
    Returns
    -------
    samples : Tensor[n_samples, D]  in the requested dtype/device
    """
    bounds = torch.as_tensor(param_bounds, dtype=dtype, device=device)
    low, high = bounds[:, 0], bounds[:, 1]
    engine = SobolEngine(dimension=len(bounds), scramble=scramble, seed=seed)
    u = engine.draw(n_samples).to(dtype=dtype, device=device)           # ∈ [0,1)
    return low + (high - low) * u                                       # rescale


# ──────────────────────────────────────────────────────────────
# 2. Serial (memory‑friendly) Sobol' swarm wrapper
# ──────────────────────────────────────────────────────────────
def sobol_swarm_opt(lm_fn,                    # your lm_direct
                    param_bounds, 
                    n_particles,
                    *,
                    dtype      = torch.float64,
                    device     = "cpu",
                    lm_kwargs  = None,        # extra kwargs to lm_fn
                    verbose    = True,
                    seed       = 42):
    """
    Runs `lm_fn` once for each Sobol‑initialised particle.
    Only one particle lives in memory at any moment.
    
    Returns
    -------
    best_X   : Tensor[D]      – parameters of the best run
    best_val : float or Tensor (scalar) – objective value (e.g. χ²) of the best run
    history  : list of dicts  – log of every particle
    """
    if lm_kwargs is None:
        lm_kwargs = {}

    # 2‑a. Generate Sobol' particles
    particles = sobol_sample(param_bounds, n_particles, 
                             dtype=dtype, device=device, seed = 16)
    
    best_val = torch.tensor(float("inf"), dtype=dtype, device=device)
    best_X   = None
    best_L   = None
    history  = []

    # 2‑b. Serial optimisation loop
    for i, x0 in enumerate(particles, 1):
        # Important: clone so lm_fn can modify the tensor safely
        result = lm_fn(x0.clone(), **lm_kwargs)
        X_opt, L_opt, chi2_opt = result                  # your lm_direct returns these

        # Log & book‑keeping
        history.append(dict(idx=i, X=X_opt.detach().cpu(), 
                            chi2=float(chi2_opt), L=float(L_opt)))
        if chi2_opt < best_val:
            best_val = chi2_opt.detach()
            best_X   = X_opt.detach().clone()
            best_L   = L_opt

        if verbose:
            print(f"[{i:>3}/{n_particles}] χ²={chi2_opt.item():.4g}   "
                  f"best={best_val.item():.4g}")

    return best_X, best_val, best_L, history



# ──────────────────────────────────────────────────────────────
# 3. Serial (memory‑friendly) Sobol' swarm wrapper
# ──────────────────────────────────────────────────────────────
def batch_chi2(points, *, Y, f, Cinv, batch_size=128):
    """
    Compute χ² for every point in `points` without building grads.
    Returns a 1‑D tensor of shape (N,).
    """
    dataset = TensorDataset(points)
    loader  = DataLoader(dataset, batch_size=batch_size, shuffle=False)

    chi2_all = []
    with torch.no_grad():                      # no autograd graph
        for (x,) in loader:
            fY = f(x)                          # (B, …)
            dY = Y.expand_as(fY) - fY          # broadcast if needed
            chi2 = (dY**2 * Cinv).sum(dim=tuple(range(1, dY.ndim)))  # (B,)
            chi2_all.append(chi2)
    return torch.cat(chi2_all)                 # (N,)

def iterated_chi2(points, *, Y, f, Cinv):
    """
    Compute χ² for every Sobol scout individually (no batching).
    Suitable when the forward model only accepts (D,) inputs.
    """
    chi2_list = []
    with torch.no_grad():
        for i in tqdm(range(len(points)), desc="Pruning Sobol' swarm"): 
            fY   = f(points[i])                      # same shape as Y
            dY   = Y - fY
            chi2 = (dY**2 * Cinv).sum()
            chi2_list.append(chi2)
    return torch.stack(chi2_list)            # (N_scouts,)

# ---------------------------------------------------------------------
# 3.1. Sobol + “genetic” pre‑selection + LM loop
# ---------------------------------------------------------------------
def sobol_ga_swarm_opt(
        lm_fn,                   # your lm_cg_og or lm_direct
        param_bounds,
        n_particles,
        n_chi,
        *,
        oversample   = 10,       # generate oversample × n_particles scouts
        keep_frac    = 0.10,     # keep this fraction (≈ n_particles)
        eval_bs      = 128,      # batch‑size for χ² scouting
        dtype        = torch.float64,
        device       = "cpu",
        lm_kwargs    = None,
        verbose      = True,
        seed         = 42,
):
    """
    Two‑stage optimisation:
      1. Draw oversample×n_particles Sobol points, keep the best keep_frac.
      2. Run `lm_fn` once on each survivor (serially, so memory stays low).

    Returns
    -------
    best_X, best_chi2, best_L, history   (same as sobol_swarm_opt)
    """
    if lm_kwargs is None:
        lm_kwargs = {}

    # unpack items needed to compute χ² quickly
    Y = lm_kwargs["Y"]
    f = lm_kwargs["f"]
    C = lm_kwargs.get("C", None)

    # --- build Cinv ONCE (matches what lm_fn will do) -----------------
    if C is None:
        Cinv = torch.ones_like(Y, dtype=dtype, device=device)
    elif C.ndim == 1:
        Cinv = 1.0 / C.to(dtype=dtype, device=device)
    else:
        Cinv = torch.linalg.inv(C.to(dtype=dtype, device=device))

    # ------------------------------------------------------------------
    # Stage 1 – oversampled scouting
    # ------------------------------------------------------------------
    n_scouts = int(math.ceil(oversample * n_particles))
    engine   = SobolEngine(dimension=len(param_bounds), scramble=True, seed=seed)

    bounds = torch.as_tensor(param_bounds, dtype=dtype, device=device)
    low, high = bounds[:, 0], bounds[:, 1]

    scouts_u = engine.draw(n_scouts).to(dtype=dtype, device=device)
    scouts   = low + (high - low) * scouts_u       # (n_scouts, D)

    # fast χ² evaluation (no grads)
    chi2_scouts = iterated_chi2(
        scouts, Y=Y.to(device), f=f, Cinv=Cinv,
    )

    # keep the best ‘keep_frac’ fraction
    k = max(1, int(keep_frac * n_scouts))
    top_vals, top_idx = torch.topk(-chi2_scouts, k)  # negative → ascending
    survivors = scouts[top_idx]                      # (k, D)

    if verbose:
        best_raw = (-top_vals).min().item()
        print(f"⇢ Scouted {n_scouts} points, best raw χ² = {best_raw/n_chi:.4g}")
        print(f"⇢ Keeping {k} best points for LM refinement …")
        print(top_vals.cpu().numpy()/n_chi)

    # ------------------------------------------------------------------
    # Stage 2 – serial LM refinement
    # ------------------------------------------------------------------
    history  = []
    best_val = torch.tensor(float("inf"), dtype=dtype, device=device)
    best_X   = None
    best_L   = None

    for i, x0 in enumerate(survivors, 1):
        res = lm_fn(x0.clone(), **lm_kwargs)           # (X, L, chi2)
        X_opt, L_opt, chi2_opt = res

        history.append(dict(idx=i,
                            X=X_opt.detach().cpu(),
                            chi2=float(chi2_opt),
                            L=float(L_opt)))

        if chi2_opt < best_val:
            best_val = chi2_opt.detach()
            best_X   = X_opt.detach().clone()
            best_L   = L_opt

        if verbose:
            print(f"[{i:>3}/{k}] χ²={chi2_opt.item():.4g}   "
                  f"best={best_val.item():.4g}")

    return best_X, best_val, best_L, history

def sobol_ga_swarm_no_nan(
        lm_fn,                   # your lm_cg_og or lm_direct
        param_bounds,
        n_particles,
        n_chi,
        *,
        oversample   = 10,       # generate oversample × n_particles scouts
        keep_frac    = 0.10,     # keep this fraction (≈ n_particles)
        eval_bs      = 128,      # batch‑size for χ² scouting
        dtype        = torch.float64,
        device       = "cpu",
        lm_kwargs    = None,
        verbose      = True,
        seed         = 42,
):
    """
    Two‑stage optimisation:
      1. Draw oversample×n_particles Sobol points, keep the best keep_frac.
      2. Run `lm_fn` once on each survivor (serially, so memory stays low).

    Returns
    -------
    best_X, best_chi2, best_L, history   (same as sobol_swarm_opt)
    """
    if lm_kwargs is None:
        lm_kwargs = {}

    # unpack items needed to compute χ² quickly
    Y = lm_kwargs["Y"]
    f = lm_kwargs["f"]
    C = lm_kwargs.get("C", None)

    # --- build Cinv ONCE (matches what lm_fn will do) -----------------
    if C is None:
        Cinv = torch.ones_like(Y, dtype=dtype, device=device)
    elif C.ndim == 1:
        Cinv = 1.0 / C.to(dtype=dtype, device=device)
    else:
        Cinv = torch.linalg.inv(C.to(dtype=dtype, device=device))

    # ------------------------------------------------------------------
    # Stage 1 – oversampled scouting
    # ------------------------------------------------------------------
    n_scouts = int(math.ceil(oversample * n_particles))
    engine   = SobolEngine(dimension=len(param_bounds), scramble=True, seed=seed)

    bounds = torch.as_tensor(param_bounds, dtype=dtype, device=device)
    low, high = bounds[:, 0], bounds[:, 1]

    scouts_u = engine.draw(n_scouts).to(dtype=dtype, device=device)
    scouts   = low + (high - low) * scouts_u       # (n_scouts, D)

    # fast χ² evaluation (no grads)
    chi2_scouts = iterated_chi2(
        scouts, Y=Y.to(device), f=f, Cinv=Cinv,
    )

    # --- Start of Patch ---
    # Filter out NaNs from the scouts' chi2 values
    valid_mask = ~torch.isnan(chi2_scouts)
    valid_chi2_scouts = chi2_scouts[valid_mask]
    valid_scouts = scouts[valid_mask]

    if verbose and not torch.all(valid_mask):
        print(f"⇢ Filtered out {torch.sum(~valid_mask)} NaN values.")

    # keep the best ‘keep_frac’ fraction from the valid scouts
    k = max(1, int(keep_frac * n_scouts))
    
    # Ensure k is not larger than the number of valid scouts
    num_valid_scouts = valid_scouts.shape[0]
    k = min(k, num_valid_scouts)

    if k > 0:
        top_vals, top_idx = torch.topk(-valid_chi2_scouts, k)  # negative → ascending
        survivors = valid_scouts[top_idx]                      # (k, D)
    else:
        # Handle the case where there are no valid scouts left
        survivors = torch.empty((0, scouts.shape[1]), dtype=dtype, device=device)
    # --- End of Patch ---

    if verbose:
        if k > 0:
            best_raw = (-top_vals).min().item()
            print(f"⇢ Scouted {n_scouts} points, found {num_valid_scouts} valid, best raw χ² = {best_raw/n_chi:.4g}")
            print(f"⇢ Keeping {k} best points for LM refinement …")
            print((-top_vals).cpu().numpy()/n_chi)
        else:
            print(f"⇢ Scouted {n_scouts} points, but no valid (non-NaN) points were found.")


    # ------------------------------------------------------------------
    # Stage 2 – serial LM refinement
    # ------------------------------------------------------------------
    history  = []
    best_val = torch.tensor(float("inf"), dtype=dtype, device=device)
    best_X   = None
    best_L   = None

    for i, x0 in enumerate(survivors, 1):
        res = lm_fn(x0.clone(), **lm_kwargs)           # (X, L, chi2)
        X_opt, L_opt, chi2_opt = res

        history.append(dict(idx=i,
                            X=X_opt.detach().cpu(),
                            chi2=float(chi2_opt),
                            L=float(L_opt)))

        if chi2_opt < best_val:
            best_val = chi2_opt.detach()
            best_X   = X_opt.detach().clone()
            best_L   = L_opt

        if verbose:
            print(f"[{i:>3}/{k}] χ²={chi2_opt.item():.4g}   "
                  f"best={best_val.item():.4g}")

    return best_X, best_val, best_L, history


import torch
from caskade import Module, forward, Param
from torch import vmap
import caustics
from caustics.light import Pixelated
from torch.nn.functional import avg_pool2d, conv2d


class CubeLens(Module):
    def __init__(
        self,
        lens,
        source_cube,
        pixelscale_source,
        pixelscale_lens,
        pixels_x_source,
        pixels_x_lens,
        upsample_factor,
        name: str = "sim",
    ):
        super().__init__(name)

        self.lens = lens
        self.source_cube = source_cube
        self.device = source_cube.device
        self.dtype = source_cube.dtype
        self.upsample_factor = upsample_factor
        self.src = Pixelated(name="source", shape=(pixels_x_source, pixels_x_source), pixelscale=pixelscale_source, image = torch.zeros((pixels_x_source, pixels_x_source)))

        # Create the high-resolution grid
        thx, thy = caustics.utils.meshgrid(
            pixelscale_lens / upsample_factor,
            upsample_factor * pixels_x_lens,
            dtype=torch.float32, device = source_cube.device, dtype = source_cube.dtype
        )

        self.thx = thx
        self.thy = thy

    @forward
    def forward(self, lens_source = True):
        cube = self.source_cube.forward()
        bx, by = self.lens.raytrace(self.thx, self.thy)

        def lens_channel(image):
            if lens_source:
                return self.src.brightness(bx, by, image = image)
            else:
                return self.src.brightness(self.thx, self.thy, image = image)
        
        # Ray-trace to get the lensed positions
        lensed_cube = vmap(lens_channel)(cube)
        del cube

        # Downsample to the desired resolution
        lensed_cube = avg_pool2d(lensed_cube[:, None], self.upsample_factor)[:, 0]
        torch.cuda.empty_cache()
        return lensed_cube


import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colormaps
from matplotlib.patches import Ellipse
from scipy.stats import norm

def corner_plot(
    chain,
    labels=None,
    figsize=(10, 10),
    true_values=None,
    ellipse_colors="g",
):
    num_params = chain.shape[1]
    cov_matrix = np.cov(chain, rowvar=False)
    mean = np.mean(chain, axis=0)
    fig, axes = plt.subplots(num_params, num_params, figsize=figsize)
    plt.subplots_adjust(wspace=0.0, hspace=0.0)
    plt.style.use('ggplot')

    for i in range(num_params):
        for j in range(num_params):
            ax = axes[i, j]

            if i == j:
                x = np.linspace(
                    mean[i] - 3 * np.sqrt(cov_matrix[i, i]),
                    mean[i] + 3 * np.sqrt(cov_matrix[i, i]),
                    100,
                )
                y = norm.pdf(x, mean[i], np.sqrt(cov_matrix[i, i]))
                ax.plot(x, y, color="g")
                ax.set_xlim(
                    mean[i] - 3 * np.sqrt(cov_matrix[i, i]),
                    mean[i] + 3 * np.sqrt(cov_matrix[i, i]),
                )
                if true_values is not None:
                    ax.axvline(true_values[i], color="red", linestyle="-", lw=1)
            elif j < i:
                ax.scatter(chain[:, j], chain[:, i], color="c", s=0.1, zorder=0)
                cov = cov_matrix[np.ix_([j, i], [j, i])]
                lambda_, v = np.linalg.eig(cov)
                lambda_ = np.sqrt(lambda_)
                angle = np.rad2deg(np.arctan2(v[1, 0], v[0, 0]))
                for k in [1, 2]:
                    ellipse = Ellipse(
                        xy=(mean[j], mean[i]),
                        width=lambda_[0] * k * 2,
                        height=lambda_[1] * k * 2,
                        angle=angle,
                        edgecolor=ellipse_colors,
                        facecolor="none",
                    )
                    ax.add_artist(ellipse)

                # Set axis limits
                margin = 3
                ax.set_xlim(
                    mean[j] - margin * np.sqrt(cov_matrix[j, j]),
                    mean[j] + margin * np.sqrt(cov_matrix[j, j]),
                )
                ax.set_ylim(
                    mean[i] - margin * np.sqrt(cov_matrix[i, i]),
                    mean[i] + margin * np.sqrt(cov_matrix[i, i]),
                )

                if true_values is not None:
                    ax.axvline(true_values[j], color="red", linestyle="-", lw=1)
                    ax.axhline(true_values[i], color="red", linestyle="-", lw=1)

            if j > i:
                ax.axis("off")

            if i < num_params - 1:
                ax.set_xticklabels([])
            else:
                if labels is not None:
                    ax.set_xlabel(labels[j])
            ax.yaxis.set_major_locator(plt.NullLocator())

            if j > 0:
                ax.set_yticklabels([])
            else:
                if labels is not None:
                    ax.set_ylabel(labels[i])
            ax.xaxis.set_major_locator(plt.NullLocator())

    plt.show()


import numpy as np
from supermage.utils.uv_utils import generate_binned_data, generate_binned_counts, generate_uv_mask, gaussian_pb
from supermage.utils.coord_utils import pixel_size_background


def dataloader_var_from_bin(u_file, v_file, sigma_squared_inv_file, npix, paduv, frequencies_file, dish_diameter):
    data_u = np.fromfile(u_file)
    data_v = np.fromfile(v_file)
    data_var = 1/np.fromfile(sigma_squared_inv_file)

    var = data_var[0::2] + 1j*data_var[1::2]
    mask_ints, deltal, deltam = generate_uv_mask(data_u, data_v, shape=(npix, npix), pad_uv=paduv)
    mask = mask_ints > 0
    fov = deltam * npix

    # Binned variances
    binned_var_real = generate_binned_data(data_u, data_v, var.real, nyquist=False, shape=(npix, npix), pad_uv=paduv)
    binned_var_imag = generate_binned_data(data_u, data_v, var.imag, nyquist=False, shape=(npix, npix), pad_uv=paduv)
    binned_counts = generate_binned_counts(data_u, data_v, nyquist=False, shape=(npix, npix), pad_uv=paduv)
    binned_var_real = binned_var_real / binned_counts
    binned_var_imag = binned_var_imag / binned_counts
    binned_var_real = binned_var_real[mask]
    binned_var_imag = binned_var_imag[mask]

    frequencies = np.load(frequencies_file)
    primary_beams = np.zeros((frequencies.shape[0], npix, npix))
    for i in range(frequencies.shape[0]):
        primary_beams[i], _ = generate_pb(diameter=dish_diameter, freq=frequencies[i], shape=(npix, npix), deltal=deltal)

    return deltam, fov, binned_var_real, binned_var_imag, mask, primary_beams


#!/usr/bin/env python

import glob
import os
import sys
from shutil import rmtree

from setuptools import Command, find_packages, setup
from setuptools.command.test import test as TestCommand


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()


install_requires = read("requirements.txt").splitlines()

# Dynamically determine extra dependencies
extras_require = {}
extra_req_files = glob.glob("requirements-*.txt")
for extra_req_file in extra_req_files:
    name = os.path.splitext(extra_req_file)[0].replace("requirements-", "", 1)
    extras_require[name] = read(extra_req_file).splitlines()

if extras_require:
    extras_require["all"] = sorted({x for v in extras_require.values() for x in v})

# Import meta data from __meta__.py (no version here)
meta = {}
exec(read("supermage/__meta__.py"), meta)

# README handling
possible_readme_names = ["README.rst", "README.md", "README.txt", "README"]
long_description = meta["description"]
readme_fname = ""
for fname in possible_readme_names:
    try:
        long_description = read(fname)
    except IOError:
        continue
    else:
        readme_fname = fname
        break

readme_ext = os.path.splitext(readme_fname)[1]
if readme_ext.lower() == ".rst":
    long_description_content_type = "text/x-rst"
elif readme_ext.lower() == ".md":
    long_description_content_type = "text/markdown"
else:
    long_description_content_type = "text/plain"


class PyTest(TestCommand):
    """Support setup.py test."""

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        pytest.main(self.test_args)


class UploadCommand(Command):
    """Support setup.py upload (local convenience)."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self): pass
    def finalize_options(self): pass

    def run(self):
        try:
            self.status("Removing previous builds...")
            here = os.path.abspath(os.path.dirname(__file__))
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution...")
        os.system(f"{sys.executable} -m build")

        self.status("Uploading the package to PyPI via Twine...")
        os.system("twine upload dist/*")

        # NOTE: We do NOT create/push tags here anymore.
        # Version is derived from existing git tags via setuptools_scm.
        sys.exit()


setup(
    name=meta["name"],
    # Version comes from git tags via setuptools_scm:
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    package_dir={meta["name"]: os.path.join(".", meta["path"])},
    python_requires=">=3.8",
    install_requires=install_requires,
    extras_require=extras_require,
    # Metadata to display on PyPI
    author=meta["author"],
    author_email=meta["author_email"],
    description=meta["description"],
    long_description=long_description,
    long_description_content_type=long_description_content_type,
    license=meta["license"],
    url=meta["url"],
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    cmdclass={
        "test": PyTest,
        "upload": UploadCommand,
    },
)




import torch, math, torch.nn.functional as F
from caskade import Module, forward, Param
import numpy as np
from supermage.utils.uv_utils import gaussian_pb
from supermage.utils.cube_tools import create_velocity_grid_stable

class VisibilityCubePadded(Module):
    """
    Identical public API except that padding is done *after* PB‑weighting.
    """
    def __init__(
        self,
        cube_simulator,
        mask,
        freqs,
        npix,                 # final grid side
        pixelscale,           # ″ / pix on the final grid
        dish_diameter: float = 12.0,
        line = "co21"
    ):
        super().__init__()
        self.cube_simulator = cube_simulator
        self.mask           = mask
        self.freqs          = freqs
        self.npix           = npix
        self.pixelscale     = pixelscale
        self.dish_diameter  = dish_diameter

        self.device = cube_simulator.device
        self.dtype  = cube_simulator.dtype
        self.flux   = Param("flux", None)

        # ── size of the *small* cube returned by cube_simulator ─────────
        self.small_side = cube_simulator.N_pix      # uses the provided attribute
        if self.small_side > self.npix:
            raise ValueError(
                f"cube_simulator.N_pix ({self.small_side}) > npix ({self.npix})."
            )
        if (self.small_side%2) != (self.npix%2):
            raise ValueError(
                "Parity mismatch between cubes! Make shapes the same parity"
            )
        # symmetric padding widths: (left,right,top,bottom)
        pad_tot   = self.npix - self.small_side
        self.pad  = (pad_tot // 2, pad_tot - pad_tot // 2) * 2  # (L,R,T,B)

        # ── primary beams ON THE SMALL GRID ─────────────────────────────
        #   (no need to generate values we’ll pad with zeros later)
        self.pb_small = torch.stack(
            [
                gaussian_pb(
                    diameter=self.dish_diameter,
                    freq=f,
                    shape=(self.small_side, self.small_side),
                    deltal=self.pixelscale,          # same pixel scale
                    device=self.device,
                    dtype=self.dtype,
                )[0]                                # gaussian_pb returns (pb, _)
                for f in freqs
            ],
            dim=0,                                  # (N_chan, S, S)
        )
        
        vel_axis, dv = create_velocity_grid_stable(f_start = freqs[0], f_end = freqs[-1], num_points = len(freqs), target_dtype = self.dtype, line = line)

        self.dv = dv[0]

    # ---------------------------------------------------------------------
    @forward
    def forward(self, plot = True, flux = None):
        """
        If `plot` is True → full UV cube masked by `mask`.
        If `plot` is False → only values where mask == True.
        """
        cube_small = self.cube_simulator.forward()          # (N_chan, S, S)

        # 1. multiply by primary beam on the same small grid
        cube_pb = cube_small * self.pb_small                # (N_chan, S, S)
        del cube_small

        # 2. scale to requested total flux *before* padding
        if flux is None:
            flux = self.flux if self.flux is not None else 1.0
        cube_pb = cube_pb * flux / self.dv / cube_pb.sum() #Integrated flux in Jy km/s

        # 3. pad *both* spatial axes to (npix, npix)
        cube_pad = F.pad(cube_pb, self.pad, mode="constant", value=0.0)
        del cube_pb

        # 4. FFT  (channel‑wise 2‑D)
        fft_cube = torch.fft.fftshift(
            torch.fft.fft2(torch.fft.ifftshift(cube_pad, dim=(-2, -1)),
                           norm="backward"),
            dim=(-2, -1),
        )
        del cube_pad                                           # memory

        # 5. deliver either the full UV plane or only the sampled points
        if plot:
            return fft_cube * self.mask.float()                # (N_chan, npix, npix)

        def gather(fft_slc, mask_slc):                         # per‑channel
            return fft_slc[mask_slc]
        return torch.vmap(gather)(fft_cube, self.mask)         # (N_chan, N_points)

class VisibilityCube(Module):
    def __init__(
        self, 
        cube_simulator, 
        #data_u, 
        #data_v, 
        #paduv, 
        mask,
        freqs, 
        npix,
        pixelscale,
        dish_diameter=12
    ):
        super().__init__()
        self.cube_simulator = cube_simulator
        #self.data_u = data_u
        #self.data_v = data_v
        #self.paduv = paduv
        self.mask = mask
        self.freqs = freqs
        self.dish_diameter = dish_diameter
        self.npix = npix
        self.pixelscale = pixelscale
        self.flux = Param("flux", None)
        self.device = cube_simulator.device
        self.dtype = cube_simulator.dtype

        #pixelscale_rad = self.pixelscale * (np.pi / 180.0) / 3600.0  # arcsec -> radians
        #self.pixel_area_sr = pixelscale_rad**2

        # Create primary beams
        pbs = []
        for freq in freqs:
            pb, _ = gaussian_pb(
                diameter=dish_diameter,
                freq=freq,
                shape=(npix, npix),
                deltal=pixelscale,
                device=self.device,
                dtype = self.dtype
            )
            pbs.append(pb)
        self.primary_beams = torch.stack(pbs, dim=0)  # shape (N_freq, Nx, Ny)

    @forward
    def forward(self, plot = True, flux = None):
        cube = self.cube_simulator.forward()

        def pb_cube(x_slice, pb_slice):
            return x_slice*pb_slice

        def fft_channel(x_slice):
            # x_slice, pb_slice: each shape (Nx, Ny)
            fft_result = torch.fft.fftshift(
                torch.fft.fft2(
                    torch.fft.ifftshift(x_slice), 
                    norm="forward")
                )
            return fft_result

        pb_results = torch.vmap(pb_cube)(cube, self.primary_beams)
        del cube
        torch.cuda.empty_cache()
        #Access velocity shifter, freqs and recalculate velocities in order to get velocity channel width
        normed_result = flux*pb_results/pb_results.sum()
        fft_results = torch.vmap(fft_channel)(normed_result)
        del normed_result
        torch.cuda.empty_cache()
        # fft_results: shape (N_freq, Nx, Ny), dtype=complex64
    
        if plot:
            # Simply multiply by the mask (broadcasting if needed)
            mask_f = self.mask.float()  # shape (N_chan, Nx, Ny)
            return fft_results * mask_f
        else:
            # We want only the masked values. Let's define a small helper 
            # that gathers the complex values from fft_results where mask is True.
            def gather_masked(fft_slc, mask_slc):
                # fft_slc: complex (Nx, Ny)
                # mask_slc: bool (Nx, Ny)
                # gather real + i imag at the True locations
                return fft_slc[mask_slc]  # shape (#Trues,)

            # We'll vmap over (N_chan) as well
            # This will return shape (N_chan, #Trues) if each channel has the same #Trues
            visibilities_result = torch.vmap(gather_masked)(fft_results, self.mask)
            # shape: (N_chan, #Trues)
            return visibilities_result



import torch
from caustics import Module, forward, Param

class ExponentialDisk2D(Module):
    """
    Purely radial exponential surface–brightness profile for a razor-thin disk.
    """
    def __init__(self):
        super().__init__()
        self.scale = Param("scale", None)              # pc

    @forward
    def brightness(self, R_map, scale=None):
        """
        I(R) ∝ exp(−R / scale)
        """
        return torch.exp(-R_map / scale)               # (H,W)


class ExponentialToroid(Module): #Formerly ExponentialUnnorm
    def __init__(self):
        super().__init__()
        self.sigma = Param("sigma", None)
        self.radius = Param("radius", None)
        self.sigma_z = Param("sigma_z", None)
        self.mu_z = Param("mu_z", None)

    @forward
    def brightness(self, x, y, z, sigma=None, radius=None, sigma_z=None, mu_z=None):
        r = torch.sqrt(x**2 + y**2)
        # phi = torch.atan2(y, x)  # Not used in calculation
        intensity = torch.exp(-0.5*(r - radius)**2 / sigma**2 - 0.5*(z - mu_z)**2 / sigma_z**2)
        return intensity


class ExponentialDisk3D(Module):
    """
    A 3D exponential disk intensity profile model, similar to KinMS.sb_profs.expdisk.
    Parameters:
    - I0: Central intensity
    - scale: Radial scale length
    - sigma_z: Vertical dispersion
    - mu_z: Vertical mean offset
    - thindisk: Boolean, if true, overrides thick disk assumption
    """

    def __init__(self, thindisk):
        super().__init__()

        # Register parameters
        self.scale = Param("scale", None)
        self.sigma_z = Param("sigma_z", None)
        self.mu_z = Param("mu_z", None)
        self.thindisk = thindisk

    @forward
    def brightness(self, x, y, z, scale=None, sigma_z=None, mu_z=None):
        """
        Computes the intensity at positions (x, y, z).

        Parameters
        ----------
        x, y, z : Tensor
            Coordinates at which the intensity is evaluated.
        scale, sigma_z, mu_z : Tensors
            Model parameters provided by caskade.
        """
        r = torch.sqrt(x**2 + y**2)
        if self.thindisk:
            dz = z[0,0,1] - z[0,0,0]
            shape = torch.exp(-r/scale) * torch.exp(-torch.abs(z)/dz)
            return shape
        else:
            intensity = torch.exp(-r/scale - 0.5*(z - mu_z)**2 / sigma_z**2)
            return intensity
    
    
    
def cutoff(r, start, end, device = "cuda"):
    """
    Creates a cutoff in a surface brightness profile between two radii. 
    This is a PyTorch-compatible version of KinMS.sb_profs.cutoff of Davis et al. (2013)
    """
    
    # Convert all entries to PyTorch tensors
    if type(r) is not torch.tensor:
        r=torch.tensor(r, device = device)
    
    return ~((r>=start)&(r<end))